-- ------------------------------------------------------------------------------ --
--                                TradeSkillMaster                                --
--                          https://tradeskillmaster.com                          --
--    All Rights Reserved - Detailed license information included with addon.     --
-- ------------------------------------------------------------------------------ --

local TSM = select(2, ...) ---@type TSM
local Open = TSM.Mailing:NewPackage("Open") ---@type AddonPackage
local ClientInfo = TSM.LibTSMWoW:Include("Util.ClientInfo")
local L = TSM.Include("Locale").GetTable()
local DelayTimer = TSM.LibTSMWoW:IncludeClassType("DelayTimer")
local String = TSM.LibTSMUtil:Include("Lua.String")
local TempTable = TSM.LibTSMUtil:Include("BaseType.TempTable")
local Money = TSM.LibTSMUtil:Include("UI.Money")
local ItemString = TSM.LibTSMTypes:Include("Item.ItemString")
local ChatMessage = TSM.Include("Util.ChatMessage")
local Theme = TSM.Include("Util.Theme")
local DefaultUI = TSM.LibTSMWoW:Include("UI.DefaultUI")
local Threading = TSM.LibTSMTypes:Include("Threading")
local ItemInfo = TSM.LibTSMService:Include("Item.ItemInfo")
local Inbox = TSM.LibTSMWoW:Include("API.Inbox")
local Mail = TSM.LibTSMService:Include("Mail")
local private = {
	settings = nil,
	thread = nil,
	isOpening = false,
	lastCheck = nil,
	moneyCollected = 0,
	checkInboxTimer = nil,
}
local INBOX_SIZE = ClientInfo.IsRetail() and 100 or 50
local MAIL_REFRESH_TIME = ClientInfo.IsRetail() and 15 or 60



-- ============================================================================
-- Module Functions
-- ============================================================================

function Open.OnInitialize(settingsDB)
	private.settings = settingsDB:NewView()
		:AddKey("global", "mailingOptions", "inboxMessages")
		:AddKey("global", "mailingOptions", "keepMailSpace")
	private.thread = Threading.New("MAIL_OPENING", private.OpenMailThread)
	private.checkInboxTimer = DelayTimer.New("MAILING_OPEN_CHECK_INBOX", private.CheckInbox)
	DefaultUI.RegisterMailVisibleCallback(private.FrameVisibleCallback)
end

function Open.KillThread()
	Threading.Kill(private.thread)

	private.PrintMoneyCollected()
	private.isOpening = false
end

function Open.StartOpening(callback, autoRefresh, keepMoney, filterText, filterType)
	Threading.Kill(private.thread)

	private.isOpening = true
	private.moneyCollected = 0

	Threading.SetCallback(private.thread, callback)
	Threading.Start(private.thread, autoRefresh, keepMoney, filterText, filterType)
end

function Open.GetLastCheckTime()
	return private.lastCheck
end



-- ============================================================================
-- Mail Opening Thread
-- ============================================================================

function private.OpenMailThread(autoRefresh, keepMoney, filterText, filterType)
	local isLastLoop = false
	while true do
		local query = TSM.Mailing.Inbox.CreateQuery()
		query:ResetOrderBy()
			:OrderBy("index", false)
			:Or()
				:Matches("itemList", filterText)
				:Matches("subject", filterText)
			:End()
			:Select("index")

		if filterType then
			query:Equal("type", filterType)
		end

		local mails = Threading.AcquireSafeTempTable()
		for _, index in query:Iterator() do
			tinsert(mails, index)
		end
		query:Release()

		private.OpenMails(mails, keepMoney, filterType)
		TempTable.Release(mails)

		if not autoRefresh or isLastLoop then
			break
		end

		local numLeftMail, totalLeftMail = Inbox.GetNumItems()
		if totalLeftMail == numLeftMail or numLeftMail == INBOX_SIZE then
			isLastLoop = true
		end

		CheckInbox()
		Threading.Sleep(1)
	end

	private.PrintMoneyCollected()
	private.isOpening = false
end

function private.CanOpenMail()
	return not C_Mail.IsCommandPending()
end

function private.OpenMails(mails, keepMoney, filterType)
	for i = 1, #mails do
		local index = mails[i]
		Threading.WaitForFunction(private.CanOpenMail)

		local mailType = Inbox.GetMailType(index)
		local matchesFilter = (not filterType and mailType) or (filterType and filterType == mailType)
		local hasBagSpace = not Mail.GetInboxItemLink(index) or CalculateTotalNumberOfFreeBagSlots() > private.settings.keepMailSpace
		if matchesFilter and hasBagSpace then
			local _, _, _, money = Inbox.GetHeaderInfo(index)
			if not keepMoney or (keepMoney and money <= 0) then
				AutoLootMailItem(index)
				private.moneyCollected = private.moneyCollected + money

				local event = private.MailIsAutoClosed(index) and "CLOSE_INBOX_ITEM" or (private.IsCraftingOrderBuy(index) and "MAIL_SUCCESS" or nil)
				if event and Threading.WaitForEvent(event, "MAIL_FAILED") ~= "MAIL_FAILED" then
					if private.settings.inboxMessages then
						private.PrintOpenMailMessage(index)
					end
					if event == "MAIL_SUCCESS" then
						private.DeleteEmptyMail(index)
					end
				end
			end
		end
	end
end



-- ============================================================================
-- Private Helper Functions
-- ============================================================================

function private.FrameVisibleCallback(visible)
	if visible then
		private.ScheduleCheck()
	else
		private.checkInboxTimer:Cancel()
	end
end

function private.CheckInbox()
	if private.isOpening then
		private.ScheduleCheck()
		return
	end

	if not TSM.UI.MailingUI.Inbox.IsMailOpened() then
		CheckInbox()
	end
	private.ScheduleCheck()
end

function private.PrintMoneyCollected()
	if private.settings.inboxMessages and private.moneyCollected > 0 then
		ChatMessage.PrintfUser(L["Total Gold Collected: %s"], Money.ToStringForUI(private.moneyCollected))
	end
	private.moneyCollected = 0
end

function private.PrintOpenMailMessage(index)
	local _, sender, subject, money, cod, _, numItems = Inbox.GetHeaderInfo(index)
	sender = sender or "?"
	local _, _, isInvoice, isConsortium = Inbox.GetText(index)
	if isInvoice then
		-- It's an invoice
		local invoiceType, itemName, playerName, bid, _, _, ahcut, _, _, quantity = Inbox.GetInvoiceInfo(index)
		playerName = playerName or (invoiceType == "buyer" and AUCTION_HOUSE_MAIL_MULTIPLE_SELLERS or AUCTION_HOUSE_MAIL_MULTIPLE_BUYERS)
		if invoiceType == "buyer" then
			local itemLink = Mail.GetInboxItemLink(index) or "["..itemName.."]"
			ChatMessage.PrintfUser(L["Bought %sx%d for %s from %s"], itemLink, quantity, Money.ToStringExact(bid, Theme.GetColor("FEEDBACK_RED"):GetTextColorPrefix()), playerName)
		elseif invoiceType == "seller" then
			ChatMessage.PrintfUser(L["Sold [%s]x%d for %s to %s"], itemName, quantity, Money.ToStringExact(bid - ahcut, Theme.GetColor("FEEDBACK_GREEN"):GetTextColorPrefix()), playerName)
		end
	elseif isConsortium then
		local reason, commission, customer, crafter = Inbox.GetCraftingOrderInfo(index)
		if reason == Enum.RcoCloseReason.RcoCloseCrafterFulfill then
			ChatMessage.PrintfUser(L["Crafting Order sold for %s to %s"], Money.ToStringExact(commission, Theme.GetColor("FEEDBACK_GREEN"):GetTextColorPrefix()), customer)
		elseif reason == Enum.RcoCloseReason.RcoCloseFulfill then
			ChatMessage.PrintfUser(L["Crafting Order bought for %s from %s"], Money.ToStringExact(commission, Theme.GetColor("FEEDBACK_RED"):GetTextColorPrefix()), crafter)
		end
	elseif numItems then
		local itemLink
		local quantity = 0
		for i = 1, numItems do
			local link = GetInboxItemLink(index, i)
			itemLink = itemLink or link
			quantity = quantity + (select(4, GetInboxItem(index, i)) or 0)
			if ItemString.Get(itemLink) ~= ItemString.Get(link) then
				itemLink = L["Multiple Items"]
				quantity = -1
				break
			end
		end
		if numItems == 1 then
			itemLink = Mail.GetInboxItemLink(index) or itemLink
		end
		local itemName = ItemInfo.GetName(itemLink) or "?"
		local itemDesc = (quantity > 0 and format("%sx%d", itemLink, quantity)) or (quantity == -1 and "Multiple Items") or "?"
		if numItems == 1 and itemLink and strfind(subject, "^" .. String.Escape(format(AUCTION_EXPIRED_MAIL_SUBJECT, itemName))) then
			ChatMessage.PrintfUser(L["Your auction of %s expired"], itemDesc)
		elseif numItems == 1 and quantity > 0 and (subject == format(AUCTION_REMOVED_MAIL_SUBJECT.."x%d", itemName, quantity) or subject == format(AUCTION_REMOVED_MAIL_SUBJECT, itemName)) then
			ChatMessage.PrintfUser(L["Cancelled auction of %sx%d"], itemLink, quantity)
		elseif cod > 0 then
			ChatMessage.PrintfUser(L["%s sent you a COD of %s for %s"], sender, Money.ToStringExact(cod, Theme.GetColor("FEEDBACK_RED"):GetTextColorPrefix()), itemDesc)
		elseif money > 0 then
			ChatMessage.PrintfUser(L["%s sent you %s and %s"], sender, itemDesc, Money.ToStringExact(money, Theme.GetColor("FEEDBACK_GREEN"):GetTextColorPrefix()))
		else
			ChatMessage.PrintfUser(L["%s sent you %s"], sender, itemDesc)
		end
	elseif money > 0 then
		ChatMessage.PrintfUser(L["%s sent you %s"], sender, Money.ToStringExact(money, Theme.GetColor("FEEDBACK_GREEN"):GetTextColorPrefix()))
	elseif subject then
		ChatMessage.PrintfUser(L["%s sent you a message: %s"], sender, subject)
	end
end

function private.IsCraftingOrderBuy(index)
	return Inbox.GetCraftingOrderInfo(index) == Enum.RcoCloseReason.RcoCloseFulfill
end

function private.MailIsAutoClosed(index)
	local isInvoice = select(3, Inbox.GetText(index))
	if isInvoice and Inbox.GetInvoiceInfo(index) == "seller_temp_invoice" then
		-- Temporary invoices are not auto deleted
		return false
	end
	local textCreated = select(8, Inbox.GetHeaderInfo(index))
	return textCreated and not private.IsCraftingOrderBuy(index)
end

function private.DeleteEmptyMail(index)
	local _, _, _, money, _, _, itemCount = Inbox.GetHeaderInfo(index)
	-- Only force delete completely empty mails
	if money == 0 and not itemCount then
		DeleteInboxItem(index)
	end
end


-- ============================================================================
-- Event Handlers
-- ============================================================================

function private.ScheduleCheck()
	if not private.lastCheck or time() - private.lastCheck > (MAIL_REFRESH_TIME - 1) then
		private.lastCheck = time()
		private.checkInboxTimer:RunForTime(MAIL_REFRESH_TIME)
	else
		local nextUpdate = MAIL_REFRESH_TIME - (time() - private.lastCheck)
		private.checkInboxTimer:RunForTime(nextUpdate)
	end
end
